<?php

declare(strict_types=1);

namespace Pinpoint\CmsBuilder\Command\Cms\Block;

use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Module\Dir;
use Magento\Framework\Module\FullModuleList;
use Magento\Store\Api\StoreRepositoryInterface;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Helper\QuestionHelper;
use Symfony\Component\Console\Helper\Table;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Question\ChoiceQuestion;
use Symfony\Component\Console\Question\ConfirmationQuestion;
use Symfony\Component\Console\Question\Question;

class Create extends Command
{
    /**
     * @var InputInterface
     */
    private InputInterface $input;

    /**
     * @var OutputInterface
     */
    private OutputInterface $output;

    /**
     * @var QuestionHelper
     */
    private QuestionHelper $questionHelper;

    /**
     * @var string
     */
    private string $patchPath = '';

    /**
     * @var string
     */
    private string $templatePath = '';

    /**
     * @param StoreRepositoryInterface $storeRepository
     * @param FullModuleList $fullModuleList
     * @param Dir $moduleDir
     * @param string|null $name
     */
    public function __construct(
        private readonly StoreRepositoryInterface $storeRepository,
        private readonly FullModuleList $fullModuleList,
        private readonly Dir $moduleDir,
        string $name = null
    ) {
        parent::__construct($name);
    }

    /**
     * @return void
     */
    protected function configure(): void
    {
        $this->setName('cms:block:create');
        $this->setDescription('Create a new CMS block from the command line');

        parent::configure();
    }

    /**
     * @param InputInterface $input
     * @param OutputInterface $output
     * @return int
     * @throws LocalizedException
     */
    protected function execute(InputInterface $input, OutputInterface $output): int
    {
        $this->questionHelper = $this->getHelper('question');
        $this->input          = $input;
        $this->output         = $output;

        $data = [];
        $data['should_be_default'] = $this->blockShouldBeDefault();
        $data['stores']            = !$data['should_be_default'] ? $this->getStoreViews() : [];
        $data['title']             = $this->getTitle();
        $data['identifier']        = $this->getIdentifier();
        $data['patch_name']        = $this->getPatchName();
        $data['patch_module']      = $this->getPatchModule();
        $data['template']          = $this->getTemplate($data['patch_module']);

        $this->createPatch($data);
        $this->createTemplate($data);
        $this->printSuccessMessage();

        return 0;
    }

    /**
     * @return bool
     */
    private function blockShouldBeDefault(): bool
    {
        $question = new ConfirmationQuestion(
            'Do you want this block to show in all store views? (Y/N) ',
            false
        );

        return (bool) $this->questionHelper->ask($this->input, $this->output, $question);
    }

    /**
     * @return array
     */
    private function getStoreViews(): array
    {
        $stores = array_map(
            function ($store) {
                return $store->getName();
            },
            $this->storeRepository->getList()
        );

        if (array_key_exists('admin', $stores)) {
            unset($stores['admin']);
        }

        $question = new ChoiceQuestion(
            'Which store view would you like this block to show in?',
            $stores
        );

        $question->setMultiselect(true);

        return $this->questionHelper->ask($this->input, $this->output, $question);
    }

    /**
     * @return string
     */
    private function getTitle(): string
    {
        $question = new Question('CMS block title: ');
        return (string) $this->questionHelper->ask($this->input, $this->output, $question);
    }

    /**
     * @return string
     */
    private function getIdentifier(): string
    {
        $question = new Question('CMS block identifier: ');
        return (string) $this->questionHelper->ask($this->input, $this->output, $question);
    }

    /**
     * @return string
     */
    private function getPatchName(): string
    {
        $question = new Question('Patch name (eg. CmsBlockFooter.php): CmsBlock');
        $result   = (string) $this->questionHelper->ask($this->input, $this->output, $question);

        return sprintf('CmsBlock%s', $result);
    }

    /**
     * @return string
     * @throws LocalizedException
     */
    private function getPatchModule(): string
    {
        $modules = array_filter(
            $this->fullModuleList->getAll(),
            function ($module) {
                $dir = $this->moduleDir->getDir($module['name']);
                return str_contains($dir, 'app/code');
            }
        );

        $moduleNames = array_map(
            function ($module) {
                return $module['name'];
            },
            $modules
        );

        if (count($moduleNames) === 0) {
            throw new LocalizedException(
                __('There are no modules in app/code for the data patch to be created. Please create a module and try again')
            );
        }

        $question = new ChoiceQuestion(
            'Which module would you like the patch to be in?',
            $moduleNames
        );

        return (string) $this->questionHelper->ask($this->input, $this->output, $question);
    }

    /**
     * @param string $moduleName
     * @return string
     */
    private function getTemplate(string $moduleName): string
    {
        $question = new Question((string) __('Template: %1::', $moduleName));
        return (string) $this->questionHelper->ask($this->input, $this->output, $question);
    }

    /**
     * @param array $data
     * @return void
     * @throws LocalizedException
     */
    private function createPatch(array $data): void
    {
        $moduleDir = $this->moduleDir->getDir($data['patch_module']);
        $patchDir  = sprintf('%s/Setup/Patch/Data', $moduleDir);

        if (!is_dir($patchDir) && !mkdir($patchDir, 0755, true)) {
            throw new LocalizedException(
                __('Could not create directory: %1', $patchDir)
            );
        }

        $this->patchPath = sprintf('%s/%s', $patchDir, $data['patch_name']);

        if (is_file($this->patchPath)) {
            throw new LocalizedException(
                __('Patch already exists: %1', $this->patchPath)
            );
        }

        if (file_put_contents($this->patchPath, $this->replaceTemplateVars($data)) === false) {
            throw new LocalizedException(
                __('Could not write to file: %1', $this->patchPath)
            );
        }
    }

    /**
     * @param array $data
     * @return string
     * @throws LocalizedException
     */
    private function replaceTemplateVars(array $data): string
    {
        $templatePath = sprintf('%s/../../../Setup/Patch/Data/CmsBlockPatch.php.template', __DIR__);
        $template     = file_get_contents($templatePath);

        if ($template === false) {
            throw new LocalizedException(
                __('Cannot read template file: %1', $templatePath)
            );
        }

        [$vendor, $module] = explode('_', $data['patch_module']);

        $vars = [
            '__VENDOR__'           => $vendor,
            '__MODULE__'           => $module,
            '__CLASS_NAME__'       => pathinfo($data['patch_name'], PATHINFO_FILENAME),
            '__STORES__'           => $this->getFormattedStores($data['stores']),
            '__BLOCK_TITLE__'      => $data['title'],
            '__BLOCK_IDENTIFIER__' => $data['identifier'],
            '__TEMPLATE__'         => sprintf('%s::%s', $data['patch_module'], $data['template'])
        ];

        foreach ($vars as $key => $value) {
            $template = str_replace($key, $value, $template);
        }

        return $template;
    }

    /**
     * @param array $stores
     * @return string
     */
    private function getFormattedStores(array $stores): string
    {
        if (count($stores) === 0) {
            return '';
        }

        $stores = array_map(
            function ($store) {
                return "'$store'";
            },
            $stores
        );

        return implode(', ', $stores);
    }

    /**
     * @param array $data
     * @return void
     * @throws LocalizedException
     */
    private function createTemplate(array $data): void
    {
        $moduleDir = $this->moduleDir->getDir($data['patch_module']);

        $this->templatePath = sprintf('%s/view/adminhtml/templates/%s', $moduleDir, $data['template']);
        $templateDir = dirname($this->templatePath);

        if (!is_dir($templateDir) && !mkdir($templateDir, 0755, true)) {
            throw new LocalizedException(
                __('Could not create directory: %1', $templateDir)
            );
        }

        if (!fopen($this->templatePath, 'w')) {
            throw new LocalizedException(
                __('Could not create file: %1', $this->templatePath)
            );
        }
    }

    /**
     * @return void
     */
    private function printSuccessMessage(): void
    {
        $table = (new Table($this->output))
            ->setHeaders(['File', 'Path'])
            ->setRows([
                ['Patch', $this->patchPath],
                ['Template', $this->templatePath]
            ]);

        $this->output->writeln(__(''));
        $this->output->writeln(__('CMS block created successfully!'));
        $table->render();
    }
}
