<?php

declare(strict_types=1);

namespace Pinpoint\CmsBuilder\Setup\Patch\Data\Cms;

use Magento\Cms\Api\BlockRepositoryInterface;
use Magento\Cms\Api\Data\BlockInterface;
use Magento\Cms\Api\GetBlockByIdentifierInterface;
use Magento\Cms\Model\BlockFactory;
use Magento\Framework\App\Area;
use Magento\Framework\App\ObjectManager;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\App\State;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Exception\NoSuchEntityException;
use Magento\Framework\Setup\ModuleDataSetupInterface;
use Magento\Framework\Setup\Patch\DataPatchInterface;
use Magento\Framework\View\Element\Template;
use Magento\Framework\View\LayoutInterface;
use Magento\Store\Api\Data\StoreInterface;
use Magento\Store\Model\StoreManagerInterface;
use Throwable;

abstract class AbstractCmsBlock implements DataPatchInterface
{
    /**
     * Stores the store codes for the stores you want the CMS block to be enabled in
     * i.e. ['store_lawsons', 'store_avs']
     * If this is empty, the block will be enabled in all stores (default)
     *
     * @var array
     */
    protected array $storeCodes = [];

    /**
     * The title of the CMS block
     *
     * @var string
     */
    protected string $blockTitle = '';

    /**
     * The identifier of the CMS block
     *
     * @var string
     */
    protected string $blockId = '';

    /**
     * The template that will be used to populate the content of the CMS block
     * i.e. "Pinpoint_Theme::a/b/c.phtml"
     * This uses the "adminhtml" area in the template directory
     *
     * @var string
     */
    protected string $template = '';

    /**
     * If the CMS block already exists before the data patch is applied, should we update that block or not?
     *
     * @var bool
     */
    protected bool $shouldUpdate = true;

    /**
     * @var StoreInterface|null
     */
    private ?StoreInterface $currentStore = null;

    /**
     * @var StoreInterface[]
     */
    private array $storesToAddBlock = [];

    /**
     * @param ModuleDataSetupInterface $moduleDataSetup
     * @param BlockRepositoryInterface $blockRepository
     * @param State $state
     * @param StoreManagerInterface $storeManager
     * @param BlockFactory $blockFactory
     * @param GetBlockByIdentifierInterface $getBlockByIdentifier
     * @param ResourceConnection $resourceConnection
     */
    public function __construct(
        private readonly ModuleDataSetupInterface $moduleDataSetup,
        private readonly BlockRepositoryInterface $blockRepository,
        private readonly State $state,
        private readonly StoreManagerInterface $storeManager,
        private readonly BlockFactory $blockFactory,
        private readonly GetBlockByIdentifierInterface $getBlockByIdentifier,
        private readonly ResourceConnection $resourceConnection
    ) {
        //
    }

    /**
     * {@inheritdoc}
     */
    public function apply(): self
    {
        $this->moduleDataSetup->startSetup();

        try {
            if (count($this->storeCodes) !== 0) {
                foreach ($this->storeCodes as $storeCode) {
                    $this->setCurrentStore($storeCode);

                    if ($this->blockAlreadyExistsInStore($this->currentStore->getId())) {
                        if ($this->shouldUpdate) {
                            $this->updateBlock();
                        }
                    } else {
                        $this->storesToAddBlock[] = $this->currentStore;
                    }
                }

                if (count($this->storesToAddBlock) > 0) {
                    $this->createBlock();
                }
            } else {
                if ($this->blockAlreadyExistsInStore(0)) {
                    if ($this->shouldUpdate) {
                        $this->updateBlock();
                    }
                } else {
                    $this->createBlock();
                }
            }
        } catch (LocalizedException) {
            // fail silently, for now...
        }

        $this->moduleDataSetup->endSetup();
        return $this;
    }

    /**
     * {@inheritdoc}
     */
    public static function getDependencies(): array
    {
        return [];
    }

    /**
     * {@inheritdoc}
     */
    public function getAliases(): array
    {
        return [];
    }

    /**
     * @param string $storeCode
     * @return void
     * @throws LocalizedException
     */
    private function setCurrentStore(string $storeCode): void
    {
        $allStores = $this->storeManager->getStores(true, true);

        if (!array_key_exists($storeCode, $allStores)) {
            throw new LocalizedException(
                __('Store code %1 does not exist', $storeCode)
            );
        }

        $this->currentStore = $allStores[$storeCode];
    }

    /**
     * @param int $storeId
     * @return BlockInterface|null
     */
    private function getBlockFromStoreId(int $storeId): ?BlockInterface
    {
        try {
            return $this->getBlockByIdentifier->execute($this->blockId, $storeId);
        } catch (NoSuchEntityException) {
            return null;
        }
    }

    /**
     * @param string|int $storeId
     * @return bool
     */
    private function blockAlreadyExistsInStore(string|int $storeId): bool
    {
        $query = $this->resourceConnection->getConnection()
            ->select()
            ->from(
                ['link_table' => $this->resourceConnection->getTableName('cms_block_store')],
                ['store_id', 'row_id']
            )
            ->joinInner(
                ['block_table' => $this->resourceConnection->getTableName('cms_block')],
                'link_table.row_id = block_table.row_id'
            )
            ->where('block_table.identifier = ?', $this->blockId)
            ->where('link_table.store_id = ?', $storeId);

        return $this->resourceConnection->getConnection()->fetchRow($query) !== false;
    }

    /**
     * @return void
     * @throws LocalizedException
     */
    private function createBlock(): void
    {
        if (count($this->storeCodes) !== 0) {
            $stores = array_map(
                function ($store) {
                    /** @var StoreInterface $store */
                    return $store->getId();
                },
                $this->storesToAddBlock
            );
        } else {
            $stores = [0];
        }

        $block = $this->blockFactory->create()
            ->setTitle($this->blockTitle)
            ->setIdentifier($this->blockId)
            ->setContent($this->getBlockContent())
            ->setStores($stores)
            ->setIsActive(true);

        $this->blockRepository->save($block);
    }

    /**
     * @return void
     * @throws LocalizedException
     */
    private function updateBlock(): void
    {
        $block = $this->getBlockFromStoreId(
            count($this->storeCodes) === 0
                ? 0
                : (int) $this->currentStore->getId()
        );

        $block->setContent($this->getBlockContent());
        $this->blockRepository->save($block);
    }

    /**
     * @return string
     * @throws LocalizedException
     */
    private function getBlockContent(): string
    {
        try {
            $content = $this->state->emulateAreaCode(Area::AREA_ADMINHTML, function () {
                $layout = ObjectManager::getInstance()->get(LayoutInterface::class);
                return $layout->createBlock(Template::class)->setTemplate($this->template)->toHtml();
            });
        } catch (Throwable $e) {
            throw new LocalizedException(
                __('Could not get contents of template: %1', $this->template),
                $e
            );
        }

        return (string) $content;
    }
}
